bl_info = {
	"name": "Msp GL datafiles",
	"author": "Mikko Rasa",
	"location": "File > Export",
	"description": "Export Msp GL data",
	"category": "Import-Export" }

if "bpy" in locals():
	import imp
	for sub in "animation", "armature", "datafile", "export_animation", "export_armature", "export_camera", "export_material", "export_mesh", "export_object", "export_scene", "export_texture", "material", "mesh", "properties", "util":
		if sub in locals():
			imp.reload(locals()[sub])

import bpy
from bpy_extras.io_utils import ExportHelper

class ExportMspGLBase(ExportHelper):
	show_progress = bpy.props.BoolProperty(name="Show progress", description="Display progress indicator while exporting", default=True)

	def execute(self, context):
		exporter = self.create_exporter()
		self.prepare_exporter(exporter)
		exporter.export_to_file(context, self.filepath)
		return {"FINISHED"}

	def create_exporter(self):
		raise Exception("create_exporter must be overridden")

	def prepare_exporter(self, exporter):
		for k, v in self.as_keywords().items():
			setattr(exporter, k, v)

class ExportMspGLMeshBase(ExportMspGLBase):
	use_strips = bpy.props.BoolProperty(name="Use strips", description="Combine faces into triangle strips", default=True)
	use_degen_tris = bpy.props.BoolProperty(name="Use degen tris", description="Concatenate triangle strips with degenerate triangles", default=False)

	def draw(self, context):
		self.general_col = self.layout.column()

		col = self.layout.column()
		col.label("Triangle strips")
		col.prop(self, "use_strips")
		col.prop(self, "use_degen_tris")

class ExportMspGLMesh(bpy.types.Operator, ExportMspGLMeshBase):
	bl_idname = "export_mesh.mspgl_mesh"
	bl_label = "Export Msp GL mesh"

	filename_ext = ".mesh"

	def create_exporter(self):
		from .export_mesh import MeshExporter
		return MeshExporter()

class ExportMspGLObject(bpy.types.Operator, ExportMspGLMeshBase):
	bl_idname = "export_mesh.mspgl_object"
	bl_label = "Export Msp GL object"

	filename_ext = ".object"

	single_file = bpy.props.BoolProperty(name="Single file", description="Write all data into a single file", default=True)
	shared_resources = bpy.props.BoolProperty(name="Shared resources", description="Use global names for resource files to enable sharing", default=True)

	export_lods = bpy.props.BoolProperty(name="Export LoDs", description="Export all levels of detail", default=True)
	use_textures = bpy.props.BoolProperty(name="Use textures", description="Use textures in the exported object", default=True)

	def create_exporter(self):
		from .export_object import ObjectExporter
		return ObjectExporter()

	def draw(self, context):
		super().draw(context)

		self.general_col.prop(self, "export_lods")
		self.general_col.prop(self, "use_textures")

		col = self.layout.column()
		col.label("Files")
		col.prop(self, "single_file")
		if not self.single_file:
			col.prop(self, "shared_resources")

class ExportMspGLArmature(bpy.types.Operator, ExportMspGLBase):
	bl_idname = "export.mspgl_armature"
	bl_label = "Export Msp GL armature"

	filename_ext = ".arma"

	def create_exporter(self):
		from .export_armature import ArmatureExporter
		return ArmatureExporter()

class ExportMspGLAnimation(bpy.types.Operator, ExportMspGLBase):
	bl_idname = "export.mspgl_animation"
	bl_label = "Export Msp GL animation"

	filename_ext = ".anim"

	export_all = bpy.props.BoolProperty(name="Export all animations", description="Export all animations present on the selected objects' NLA tracks")
	collection = bpy.props.BoolProperty(name="As a collection", description="Export the animations as a single collection file", default=True)

	def check(self, context):
		result = False

		ext = ".mdc" if self.export_all and self.collection else ".anim"
		ext_changed = (ext!=self.filename_ext)
		if ext_changed:
			if self.filepath.endswith(self.filename_ext):
				self.filepath = self.filepath[:-len(self.filename_ext)]
			self.filename_ext = ext

		super_result = super().check(context)

		return ext_changed or super_result

	def create_exporter(self):
		from .export_animation import AnimationExporter
		return AnimationExporter()

	def draw(self, context):
		col = self.layout.column()
		col.prop(self, "export_all")
		if self.export_all:
			col.prop(self, "collection")

class ExportMspGLScene(bpy.types.Operator, ExportMspGLBase):
	bl_idname = "export_scene.mspgl_scene"
	bl_label = "Export Msp GL scene"

	filename_ext = ".scene"

	resource_collection = bpy.props.BoolProperty(name="Resource collection", description="Put resources to a single collection file", default=True)

	def create_exporter(self):
		from .export_scene import SceneExporter
		return SceneExporter()

	def draw(self, context):
		col = self.layout.column()
		col.prop(self, "resource_collection")

class ExportMspGLCamera(bpy.types.Operator, ExportMspGLBase):
	bl_idname = "export.mspgl_camera"
	bl_label = "Export Msp GL camera"

	filename_ext = ".camera"

	def create_exporter(self):
		from .export_camera import CameraExporter
		return CameraExporter()

def menu_func_export(self, context):
	self.layout.operator(ExportMspGLMesh.bl_idname, text="Msp GL mesh")
	self.layout.operator(ExportMspGLObject.bl_idname, text="Msp GL object")
	self.layout.operator(ExportMspGLArmature.bl_idname, text="Msp GL armature")
	self.layout.operator(ExportMspGLAnimation.bl_idname, text="Msp GL animation")
	self.layout.operator(ExportMspGLScene.bl_idname, text="Msp GL scene")
	self.layout.operator(ExportMspGLCamera.bl_idname, text="Msp GL camera")

from .properties import MspGLMeshProperties, MspGLObjectProperties

def register():
	bpy.utils.register_module(__name__)

	bpy.types.INFO_MT_file_export.append(menu_func_export)

	from .properties import register_properties
	register_properties()

def unregister():
	bpy.utils.unregister_module(__name__)

	bpy.types.INFO_MT_file_export.remove(menu_func_export)

if __name__=="__main__":
	register()
