#ifndef MSP_GL_UNIFORM_H_
#define MSP_GL_UNIFORM_H_

#include <algorithm>
#include "program.h"

namespace Msp {
namespace GL {

class Uniform
{
protected:
	Uniform() { }
private:
	Uniform(const Uniform &);
	Uniform &operator=(const Uniform &);
public:
	virtual ~Uniform() { }

	virtual void apply(int) const = 0;
	virtual void store(const Program::UniformInfo &, void *) const = 0;
	virtual Uniform *clone() const = 0;
};


template<typename T>
class UniformScalar: public Uniform
{
public:
	typedef T BaseType;
	typedef T Type;

private:
	Type value;

public:
	UniformScalar(Type v): value(v) { }

	virtual void apply(int index) const
	{ apply(index, 1, &value); }

	static void apply(int, unsigned, const T *);

	virtual void store(const Program::UniformInfo &info, void *buffer) const
	{ store(info, buffer, &value); }

	static void store(const Program::UniformInfo &, void *buffer, const T *value)
	{ *reinterpret_cast<T *>(buffer) = *value; }

	virtual UniformScalar *clone() const
	{ return new UniformScalar(value); }
};

typedef UniformScalar<int> Uniform1i;
typedef UniformScalar<float> Uniform1f;


template<typename T, unsigned vecsize>
class UniformVector: public Uniform
{
public:
	typedef T BaseType;
	typedef T Type[vecsize];

private:
	Type value;

public:
	UniformVector(const T *vp)
	{ std::copy(vp, vp+vecsize, value); }

	virtual void apply(int index) const
	{ apply(index, 1, value); }

	static void apply(int index, unsigned size, const T *value);

	virtual void store(const Program::UniformInfo &info, void *buffer) const
	{ store(info, buffer, value); }

	static void store(const Program::UniformInfo &, void *buffer, const T *value)
	{ std::copy(value, value+vecsize, reinterpret_cast<T *>(buffer)); }

	virtual UniformVector *clone() const
	{ return new UniformVector(value); }
};

typedef UniformVector<int, 2> Uniform2i;
typedef UniformVector<float, 2> Uniform2f;
typedef UniformVector<int, 3> Uniform3i;
typedef UniformVector<float, 3> Uniform3f;
typedef UniformVector<int, 4> Uniform4i;
typedef UniformVector<float, 4> Uniform4f;


template<typename T, unsigned rows, unsigned cols>
class UniformMatrix: public Uniform
{
public:
	typedef T BaseType;
	typedef T Type[rows*cols];

private:
	Type value;

public:
	UniformMatrix(const T *vp)
	{ std::copy(vp, vp+rows*cols, value); }

	virtual void apply(int index) const
	{ apply(index, 1, value); }

	static void apply(int index, unsigned size, const T *value);

	virtual void store(const Program::UniformInfo &info, void *buffer) const
	{ store(info, buffer, value); }

	static void store(const Program::UniformInfo &info, void *buffer, const T *value)
	{
		for(unsigned i=0; i<cols; ++i)
			UniformVector<T, rows>::store(info, reinterpret_cast<char *>(buffer)+i*info.matrix_stride, value+i*rows);
	}

	virtual UniformMatrix *clone() const
	{ return new UniformMatrix(value); }
};

typedef UniformMatrix<float, 2, 2> UniformMatrix2x2f;
typedef UniformMatrix<float, 3, 3> UniformMatrix3x3f;
typedef UniformMatrix<float, 4, 4> UniformMatrix4x4f;


template<typename T>
class UniformArray: public Uniform
{
private:
	typedef typename T::BaseType BaseType;
	enum { elemsize = sizeof(typename T::Type)/sizeof(typename T::BaseType) };

	BaseType *values;
	unsigned size;

public:
	UniformArray(unsigned n, const BaseType *vp):
		size(n)
	{
		values = new BaseType[elemsize*size];
		std::copy(vp, vp+elemsize*size, values);
	}

	~UniformArray()
	{
		delete[] values;
	}

	virtual void apply(int index) const
	{ T::apply(index, size, values); }

	virtual void store(const Program::UniformInfo &info, void *buffer) const
	{
		for(unsigned i=0; i<size; ++i)
			T::store(info, reinterpret_cast<char *>(buffer)+i*info.array_stride, values+i*elemsize);
	}

	virtual UniformArray *clone() const
	{ return new UniformArray(size, values); }
};

} // namespace GL
} // namespace Msp

#endif
